/*
    RawSpeed - RAW file decoder.

    Copyright (C) 2022 Roman Lebedev

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
*/

#pragma once

#include <type_traits> // for enable_if_t

namespace rawspeed {

template <class T> struct NotARational {
public:
  using value_type = T;

  T num;
  T den;

  NotARational() = default;
  NotARational(T num_, T den_) : num(num_), den(den_) {}

  template <typename T2,
            std::enable_if_t<std::is_floating_point_v<T2>, bool> = true>
  explicit operator T2() const {
    return T2(num) / T2(den);
  }
};

} // namespace rawspeed
